<?php

/**
 * Vvveb
 *
 * Copyright (C) 2022  Ziadin Givan
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 */

namespace Vvveb\Controller;

use function Vvveb\dashesToCamelCase;
use function Vvveb\humanReadable;
use function Vvveb\siteSettings;
use Vvveb\System\Cache;
use Vvveb\System\Sites;
use Vvveb\System\Sqlp\Sqlp;

#[\AllowDynamicProperties]
class Schema extends Base {
	private $models = [];

	private $tables = [];

	private $tableNames = [];

	function type($type) {
		$compare = strtolower($type);

		switch ($compare) {
			case 'int':
			case 'tinyint':
			case 'decimal':
			case 'smallint':
			case 'double':
			case 'array':
				$type = 'Int';

				break;

			case 'varchar':
			case 'char':
			case 'text':
			case 'enum':
			case 'date':
			case 'datetime':
			case 'timestamp':
			case 'longtext':
			case 'tinytext':
				$type = 'String';

				break;
		}

		return $type;
	}

	function types() {
		$db = \Vvveb\System\Db::getInstance();

		$tables  = $db->getTableNames(DB_NAME);
		$types   = [];

		foreach ($tables as $table) {
			$meta = $db->getColumnsMeta($table, true);
			$cols = [];

			foreach ($meta as $column) {
				//var_dump($column);
				$name               = $column['name'] ?? '';
				$col                = [];
				$col['name']        = $name;
				$col['description'] = $column['c'] ?? '';

				//colum is input type
				if (substr_compare($name, '_id', -3 ,3) === 0) {
					$type     = 'ID';
					$relation = substr($name,0, -3);

					if ($relation != $table && strpos($table, $relation) === 0) {
						$relationType                           =  dashesToCamelCase($table, '_') . 'Type';
						$types[$relation]['properties'][$table] = ['name' => $table, 'type' => "[$relationType]"];
					}
				} else {
					if (in_array($name, $tables)) {
						$type = dashesToCamelCase($name, '_') . 'Type'; //. 'Input';
					} else {
						$type = $this->type($column['t'] ?? 'Int');
					}
				}

				$col['type'] = $type;

				$cols[$column['name']] = $col;
			}

			$data = [
				'name'       => $table,
				'properties' => $cols,
			];

			$types[$table] = $data;

			//for content tablee also add name and content columns to relation
			if (substr_compare($table, '_content', -8 ,8) === 0) {
				$relation = substr($table, 0,  -8);
				$types[$relation]['properties']['name']    = ['type' => 'char', 'name' => 'name'];
				$types[$relation]['properties']['content'] = ['type' => 'char', 'name' => 'content'];
				$types[$relation]['properties']['slug']    = ['type' => 'char', 'name' => 'slug'];
			}
		}

		//var_export($schemas);
		return $types;
	}

	function models() {
		$dirSQL = DIR_ROOT . 'admin' . DS . 'sql' . DS . DB_ENGINE . DS;
		$sqlp   = new Sqlp();
		$models = [];

		$files = glob("$dirSQL*.sql");
		//var_dump($files);

		foreach ($files as $file) {
			$sqlp->parseSqlPfile($file);
			$tree  = $sqlp->getModel();
			$model = basename($file, '.sql');

			foreach ($tree as $method => $options) {
				$methodName = lcfirst(dashesToCamelCase($method . '_' . $model, '_'));
				//var_dump($methodName);
				$models[$model][$methodName] = $options['params'];
			}
		}

		return $models;
	}

	function schema() {
		$typesText = '';
		$types     = $this->types();

		foreach ($types as $table => $attrs) {
			$properties = '';

			foreach ($attrs['properties'] as $name => $property) {
				$type = $this->type($property['type'] ?? 'Int');
				$properties .= "  $name: $type\n";
			}

			$tableName = dashesToCamelCase($table, '_');
			$typesText .= "type {$tableName}Type {\n$properties}\n\n";
		}

		$models    = $this->models();
		$queries   = [];
		$mutations = [];

		$queriesText     = '';
		$mutationsText   = '';
		$connectionsText = '';

		foreach ($models as $model => $methods) {
			$modelComment   =  "\n\n  # " . humanReadable($model) . "\n";
			$mutationsText .= $modelComment;
			$queriesText .= $modelComment;

			if ($model == 'category' || $model == 'filter' || $model == 'stat') {
				continue;
			}

			foreach ($methods as $method => $params) {
				$parameters = '';

				foreach ($params as $name => $param) {
					if ($param['in_out'] == 'IN') {
						$name = $param['name'];
						$type = $this->type($param['type'] ?? 'Int');
						$parameters .= "$name: $type, ";
					}
				}

				$parameters = rtrim($parameters, ', ');
				$type       = dashesToCamelCase($model, '_');

				if (strpos($method, 'get') === 0) {
					if ($model == 'category' || $model == 'filter' || $model == 'stat') {
						//$type = 'Taxonomy';
					}

					$return = $type . 'Type';

					if (strpos($method, 'getAll') === 0) {
						$return = "[{$type}Type]";
					}

					$methodText = "  $method($parameters):{$return}\n";

					$queriesText .= $methodText;
				} else {
					$methodText = "  $method($parameters):ID\n";

					$mutationsText .= $methodText;
				}

				if (strpos($method, 'getAll') === 0) {
					$modelPlural = lcfirst($type) . 's';
					$methodText  = "  $modelPlural($parameters):{$type}Connection\n";
					$queriesText .= $methodText;

					$connectionsText .= $modelComment .
					"type {$type}Connection {\n  nodes:[{$type}Type!]!\n  pageInfo:PageInfo!\n}\n";
				}
			}
		}

		$cache     = Cache::getInstance();
		$json      = $cache->cache(APP,'graphql-schema',function () {
			$site = siteSettings();
			$data = Sites :: getSiteData();

			$json = [];
			$json['types'] = $this->types();

			return $json;
		}, 259200);

		$text = <<<GQL
type PageInfo {
  count: Int
  page: Int
  limit: Int
  endCursor: String
  hasNextPage: Boolean!
  hasPreviousPage: Boolean!
  startCursor: String
}
		
$connectionsText

$typesText

type RootQueryType {
$queriesText
}

type MutationType {
$mutationsText
}

schema {
  query:RootQueryType
  mutation:MutationType
}
GQL;

		return $text;
	}

	function index() {
		$text = $this->schema();
		$this->response->setType('text');
		$this->response->output($text);
	}
}
